import os
from importlib.metadata import version

from dotenv import load_dotenv
from flask import Flask
from pi_conf import Config

load_dotenv()
ROOT_DIR = os.path.dirname(os.path.abspath(__file__))


class QaiFlask(Flask):
    def __init__(self, *args, cfg=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.init(cfg)

    def init(self, cfg: Config):
        self.website_2_id: dict[str, str] = None
        self.id_2_website: dict[str, str] = None
        self.cfg: Config = cfg
        if "aws" in self.cfg and self.cfg.aws.get("enabled", True):
            from qai.storage.aws.aws_history import AWSCompanyIds

            loader = AWSCompanyIds()
            self.website_2_id = loader.website_2_id
            self.id_2_website = loader.id_2_website

        if "website_name_map" in self.cfg:
            self.website_2_id = {}
            self.website_2_id.update(self.cfg.website_name_map)
            self.id_2_website = {v: k for k, v in self.website_2_id.items()}


def create_app(_cfg: Config) -> QaiFlask:
    return QaiFlask(__name__, cfg=_cfg)


try:
    VERSION = version("qai-server")
except Exception:
    VERSION = "0.0.0"



import functools
from typing import Any

from flask import jsonify, request
from pydantic import BaseModel


class mockable:
    """
    Decorator to check for required variables in the request
    """

    def __init__(self, model_cls: type[BaseModel]):
        self.model_cls = model_cls

    def __call__(self, func):
        @functools.wraps(func)
        def decorated_function(*args, **kwargs):
            instance = self.model_cls(**request.get_json())
            is_mock = getattr(instance, "mock", False)
            if is_mock:
                ofunc = getattr(__import__('qai.server.mock.mock_functions', fromlist=[func.__name__]), func.__name__)
                return ofunc(*args, **kwargs)
            return func(*args, **kwargs)

        return decorated_function



import functools
from logging import getLogger

from flask import current_app as app
from flask import request

log = getLogger(__name__)


def token_required(func):
    """
    Decorator to check for a valid token
    """

    @functools.wraps(func)
    def decorated_function(*args, **kwargs):
        try:
            server_config = app.cfg.server
        except AttributeError:
            raise AttributeError(
                f"Error! No server config. set server.allowed_tokens in config file."
            )

        try:
            allowed_tokens = set(server_config.allowed_tokens)
        except AttributeError:
            raise AttributeError(
                f"Error! No allowed tokens in config. set server.allowed_tokens in config file. "
            )
        ## Get and remove token from request, we don't want to be printing this
        token = request.get_json().pop("token", "")
        if token not in allowed_tokens:
            log.error(f"Error! Invalid token <{token}> === {allowed_tokens}")
            raise Exception("Error! Invalid token")
        return func(*args, **kwargs)

    return decorated_function


def check_variable_exists(variables, in_data) -> bool:
    """
    Check if a variable exists in the request
    Args:
        variables: str or list of str
        in_data: dict
    Returns:
        bool : True if variable exists
    """
    if isinstance(variables, str):
        variables = [variables]
    for variable in variables:
        if variable not in in_data:
            raise Exception(f"Error! {variable} not in request")
    return True


def check_variable_one_of(variables, in_data, allow_more_than_one=False) -> bool:
    """
    Check if a variable s one of many
    """
    if isinstance(variables, str):
        variables = [variables]
    if not isinstance(variables, set):
        variables = set(variables)
    found = False
    for input_var in in_data.keys():
        if input_var in variables:
            if found:
                raise Exception(f"Error! Both {found} and {input_var} were in request")
            if allow_more_than_one:
                return True
            found = input_var
    if not found:
        raise Exception(f"Error! None of {variables} were in {in_data} request")
    return True



from flask import jsonify, request
from qai.agent.basic import BasicQuery, QueryType

from qai.server.mock.mock_responses import (
    campaign_response,
    company_response,
    people_response,
)
from qai.server.models import CampaignInputModel


def campaign():
    model = CampaignInputModel(**request.get_json())
    basic_query = BasicQuery()
    result = basic_query.query(model.query)
    if result == QueryType.campaign:
        return jsonify(campaign_response), 200
    elif result == QueryType.company:
        return jsonify(company_response), 200
    elif result == QueryType.people:
        return jsonify(people_response), 200
    else:
        return jsonify(campaign_response), 200

def company_chat():
    return {"response": "Mocked Response"}, 200


import uuid
from pprint import pformat

from flask import jsonify, request
from flask_pydantic import validate
from pi_log import logs
from qai.agent.agents.campaign_agent import CampaignAgent
from qai.chat.chat import company_query
from qai.chat.db.chroma.chroma import Chroma

from qai.server import VERSION, create_app
from qai.server.config import cfg
from qai.server.mock import mockable
from qai.server.models import CampaignInputModel, CompanyChatbotModel

## app should be created before importing other modules that require cfg
app = create_app(cfg) 

from qai.server.tools.security import token_required

log = logs.getLogger(__name__)
logs.set_app_level("debug")
default_not_found_message = "Sorry, I don't know the answer to that question."

chat_session = {}


@app.route("/heartbeat", methods=["GET", "POST"])
def heartbeat():
    log.debug(f"serve::heartbeat")
    return jsonify({"status": "OK"}), 200


@app.route("/set_config", methods=["GET", "POST"])
@token_required
def set_config(key: str, value: str):
    """
    Set a configuration value
    """
    in_data = request.get_json()
    log.debug(f"serve::set_config, in_data={in_data}")
    old = app.cfg[key]
    app.cfg[key] = value
    return jsonify({"old": old, "new": value}), 200


@app.route("/set_loglevel/<level>", methods=["GET", "POST"])
@token_required
def set_loglevel(level: str):
    """
    Set the log level
    Args:
        level: log level
    """
    logs.set_log_level(level)
    return jsonify({"status": "OK", "message": f"log level set to {level}"}), 200


@app.route("/version", methods=["GET", "POST"])
@token_required
def version():
    """
    Get the version of the app
    """
    log.debug(f"serve::version")
    return jsonify({"config_version": app.cfg.version, "app_version": VERSION}), 200


@app.route("/get_collection", methods=["GET", "POST"])
@token_required
def chroma():
    in_data = request.get_json()
    log.debug(f"serve::chroma, in_data={in_data}, chroma.cfg={app.cfg.chroma}")
    c = Chroma(app.cfg.chroma)
    col = c.get_collection(in_data["collection_name"])
    count = col.count() if col else -1
    meta = col.metadata if col else {}
    log.debug(f"serve::chroma count={count}, metadata={meta}")
    return jsonify({"count": count, "in_data": in_data, "metadata": meta}), 200


@app.route("/config", methods=["GET", "POST"])
@token_required
def config():
    log.debug(f"serve::config \n{pformat(app.cfg)}")
    return (
        jsonify(
            {
                "config": app.cfg,
            }
        ),
        200,
    )


def _debug_query(in_data, guid: str, query: str, company_name: str):
    query = query[36:].strip().lstrip(":").strip()

    in_data.pop("query")
    d = {
        "app_version": VERSION,
        "config_version": app.cfg.version,
        "build_time": app.cfg.build_time,
        "query": query,
        "in_data": in_data,
    }
    in_data["response"] = pformat(d)
    return jsonify(in_data), 200


@app.route("/", methods=["GET", "POST"])
@token_required
@mockable(CompanyChatbotModel)
@validate(body=CompanyChatbotModel)
def company_chat():
    """
    Query the company chatbot
    """
    model = CompanyChatbotModel(**request.get_json())

    log.debug(f"company_chat(), in_data={model}")

    try:
        response = company_query(
            query=model.query,
            user_id=model.user_id,
            company_name=model.company_name,
            company_id=model.company_id,
            model_cfg=app.cfg.model,
            chroma_cfg=app.cfg.chroma,
        )
    except Exception as e:
        return {"response": str(e)}, 500
    response = {"response": str(response)}
    return jsonify(response), 200


@app.route("/campaign", methods=["GET", "POST"])
@mockable(CampaignInputModel)
@validate(body=CampaignInputModel)
@token_required
def campaign():
    """
    Do a campaign query
    """
    model = CampaignInputModel(**request.get_json())
    log.debug(f"campaign() : model={model}")
    if not model.id:
        model.id = uuid.uuid4().hex

    agent: CampaignAgent
    if model.user_id in chat_session:
        agent = chat_session[model.user_id]
    else:
        agent = CampaignAgent.create(**model.to_params())

    if model.uploaded_data:
        ## convert list of dict to dict of dict
        uploaded_data = {d.get("id", uuid.uuid4()): d for d in model.uploaded_data}
        agent.people = uploaded_data
    response: CampaignResponse = agent.chat(model.query)
    js = {}
    js["actions"] = [
        {
            "action": "text",
            "response": response.response,
        }
    ]
    if response.emails:
        for em in response.emails:
            e_js = em.dict()
            e_js.update(
                {
                    "type": "email_sequence_draft",
                    "sequence_id": response.sequence_id,
                    "title": em.subject,
                }
            )
            js["actions"].append(e_js)

    return jsonify(js), 200


if __name__ == "__main__":
    app.run(host="0.0.0.0", debug=True)



campaign_response = {
    "actions": [
        {
            "type": "list",
            "title": "Here are some of the people:",
            "values": [
                {
                    "name": "John Doe",
                    "email": "johndoe@example.com",
                    "phone_number": "1234567890",
                },
                {
                    "name": "John Doe 2",
                    "email": "johndoe2@example.com",
                    "phone_number": "1234567891",
                },
            ],
        },
        {
            "type": "email_sequence_draft",
            "title": "Here is the draft",
            "subject": "Exclusive Christmas Offer: 15% Off until Jan 2, 2024!",
            "content": "Celebrate this season with our exclusive Christmas offer! Get ahead in 2024 by enjoying a 15% discount on [Your Product/Service]. Embrace the festive spirit and make the most of this limited-time offer. Act before Jan 2, 2024, to avail yourself of this special discount",
            "sequence_id": "140bb6d6-1c1d-4129-904a-2bb5e0f573d2",
        },
    ]
}

company_response = {
    "actions": [
        {
            "type": "list",
            "title": "Here are some of the companies:",
            "values": [
                {
                    "name": "Company A",
                    "email": "zz@not.an.email.com",
                    "linkedin": "https://www.linkedin.com/company/company-a",
                },
                {
                    "name": "Company B",
                    "email": "bb@not.an.email.com",
                    "linkedin": "https://www.linkedin.com/company/company-b",
                },
            ],
        }
    ]
}

people_response = {
    "actions": [
        {
            "type": "list",
            "title": "Here are some of the people:",
            "values": [
                {
                    "name": "Keerthi",
                    "email": "keerthi@not.an.email.com",
                    "phone_number": "1234567890",
                },
                {
                    "name": "Eva",
                    "email": "eva@not.an.email.com",
                    "phone_number": "1234567891",
                },
            ],
        }
    ]
}



from typing import Optional
from uuid import UUID, uuid4

import pydantic
from pydantic import BaseModel, ConfigDict, Field, ValidationError


class RequestModel(BaseModel):
    id: UUID = Field(default_factory=uuid4)
    mock: Optional[bool] = False
    token: Optional[str] = None

    def to_params(self) -> dict:
        return self.model_dump()

class CompanyChatbotModel(RequestModel):
    query: str
    user_id: str
    company_id: Optional[str] = None
    company_name: Optional[str] = None
    asynchronous: Optional[str] = False
    model: Optional[dict] = None
    uploaded_data: Optional[list] = None
    sender_company: Optional[dict] = None
    sender_person: Optional[dict] = None
    model_config = ConfigDict(extra="allow")

    @pydantic.model_validator(mode="after")
    def validate_company(self):

        ## If company_id is not provided, then company_name must be provided
        if not self.company_id and not self.company_name:
            raise ValidationError("company_id or company_name must be provided.")
        ## if both company_id and company_name are provided, then throw an error
        if self.company_id and self.company_name:
            raise ValidationError(
                f"company_id and company_name cannot both be provided. "
                f"Got {self.company_id} and {self.company_name}"
            )
        return self

    def _fill_company_info(self, website_2_id: dict[str, str], id_2_website: dict[str, str]):
        if self.company_name:
            self.company_id = website_2_id.get(self.company_name)
        else:
            self.company_name = id_2_website.get(self.company_id)
        return self
            


class CampaignInputModel(RequestModel):
    query: str
    user_id: str
    session_id: Optional[str] = Field(alias="sequence_id", default=None)
    model: Optional[dict] = None
    uploaded_data: Optional[list[dict]] = None
    sender_company: Optional[dict] = None
    sender_person: Optional[dict] = None
    on_complete_emails_url: Optional[str] = Field(alias="asynchronous", default=None)

    model_config = ConfigDict(
        extra="allow",
    )

    def to_params(self):
        remove_keys = set(
            [
                "query",
                "user_id",
                "session_id",
                "company_id",
                "conversation",
            ]
        )
        remove_keys |= set(RequestModel.model_fields.keys())
        remove_keys.add("query")
        d = self.model_dump()
        people_list = d.pop("uploaded_data", None)
        if people_list:
            pdict = {}
            for p in people_list:
                pid = p.get("id")
                if not pid:
                    pid = str(uuid4())
                pdict[pid] = p
        d["on_complete_emails_url"] = d.pop("asynchronous", None)
        d["from_person"] = d.pop("sender_person", None)
        d["from_company"] = d.pop("sender_company", None)
        d["model_config"] = d.pop("model", None)

        for k in remove_keys:
            d.pop(k, None)

        return d
